package cz.zcu.mjaneck.bak_prace.communication;

import com.fazecast.jSerialComm.SerialPort;
import javafx.application.Platform;
import javafx.collections.ObservableMap;
import javafx.scene.chart.XYChart;
import cz.zcu.mjaneck.bak_prace.app.Utility;

import java.io.IOException;
import java.io.OutputStream;
import java.nio.ByteBuffer;

/**
 * Třída reprezentující vlákno pro komunikaci s Arduinem
 */
public class ArduinoCommunicationThread implements Runnable {

	/** Pole pro zaznemanávání naměřených hodnot a jejich času zaznamenání */
	private XYChart.Series<Number, Number> series;
	/** Boolean proměnná určující jestli má vlákno skončit svůj běh */
	private boolean stopFlag;
	/** Zvolená velikost proudu vybíjení */
	private int current;
	/** Zvolený port */
	private SerialPort port;
	/** Mapa kapacit */
	private ObservableMap<String, Double> capacityMap;
	/** Stream pro odesílání hodnot na Arduino */
	private OutputStream outputStream;
	/** čas začátku komunikace */
	private long startTime;
	/** Určuje zda se jedná o první iteraci while cyklu uvnitř vlákna */
	private boolean firstIteration = true;

	/**
	 * Konstruktor vlákna
	 * @param series - pole pro hodnoty a čas
	 * @param port - zvolený port
	 * @param current - zvolená velikost proudu
	 * @param capacityMap - mapa kapacit
	 */
	public ArduinoCommunicationThread(XYChart.Series<Number, Number> series, String port, int current,
									  ObservableMap<String, Double> capacityMap) {
		this.stopFlag = false;
		this.current = current;
		this.series = series;
		this.port = SerialPort.getCommPort(port);
		this.capacityMap = capacityMap;
	}

	/**
	 * Metoda run vlákna. Běží dokud není stopFlag nastaven na true.
	 */
	@Override
	public void run() {
		startTime = 0;

		port.openPort();
		port.setComPortTimeouts(SerialPort.TIMEOUT_READ_BLOCKING, 60000, 0);
		outputStream = port.getOutputStream();

		try {
			outputStream.write(current / 20);
		} catch (IOException e) {
			e.printStackTrace();
		}

		while(!stopFlag) {
			byte[] bytes = new byte[2];
			port.readBytes(bytes, 2);
			int in = ByteBuffer.wrap(Utility.expandArray(bytes)).getInt();

			if (!stopFlag) {
			    double time;

			    if(firstIteration) {
			        time = 0;
			        startTime = System.currentTimeMillis();
                } else {
			    	time = System.currentTimeMillis() - startTime;
				}

				Platform.runLater(() -> series.getData().add(
						new XYChart.Data<>(time / 1000, in)
				));

				capacityMap.put(series.getName().substring(0, series.getName().lastIndexOf('-')),
						calculateCapacity((int)time / 1000));

				firstIteration = false;
			}
		}
	}

	/**
	 *  Nastaví stopFlag na true a odesílá nulovou hodnotu na arduino.
	 */
	public void stopThread() {
		try {
			outputStream.write(0);
		} catch (IOException e) {
			e.printStackTrace();
		}
		this.stopFlag = true;
		this.port.closePort();
	}

	/**
	 * Vrací čas začátku komunikace.
	 * @return - čas začátku měření
 	 */
	public long getStartTime() {
		return startTime;
	}

	/**
	 * Vypočte hodnotu kapacity baterie.
	 * @param time - Čas zaznamenání poslední naměřené hodnoty
	 * @return - vypočtená kapacita
	 */
	private double calculateCapacity(int time) {
		return (double)(current * time) / 3600.0;
	}
}
