package cz.zcu.mjaneck.bak_prace.gui;

import com.fazecast.jSerialComm.SerialPort;
import javafx.animation.Animation;
import javafx.animation.KeyFrame;
import javafx.animation.Timeline;
import javafx.application.Application;
import javafx.application.Platform;
import javafx.beans.binding.Bindings;
import javafx.collections.FXCollections;
import javafx.collections.MapChangeListener;
import javafx.collections.ObservableList;
import javafx.collections.ObservableMap;
import javafx.geometry.Insets;
import javafx.geometry.Pos;
import javafx.scene.Node;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.SnapshotParameters;
import javafx.scene.chart.LineChart;
import javafx.scene.chart.NumberAxis;
import javafx.scene.chart.XYChart.Series;
import javafx.scene.chart.XYChart;
import javafx.scene.control.*;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.VBox;
import javafx.stage.FileChooser;
import javafx.stage.Stage;
import javafx.util.Duration;
import cz.zcu.mjaneck.bak_prace.app.DataHandler;
import cz.zcu.mjaneck.bak_prace.app.MeasurementManager;
import cz.zcu.mjaneck.bak_prace.app.Utility;

import java.io.*;
import java.util.*;

/**
 * Třída obsluhjící gui
 */
public class MeasurementWindow extends Application {

    /** Proměnná obsahující řetězcové zdroje */
    private static Properties properties;
    /** Aplikační okno */
    private Stage stage;
    /** Instance třídy pro správu měření */
    private MeasurementManager manager = MeasurementManager.getInstance();
    /** Instance třídy pro správu dat */
    private DataHandler handler = DataHandler.getInstance();

    /** Graf, ve kterém jsou zobrazena měření */
    private LineChart<Number, Number> lineChart;
    /** Mapa kapacit měření */
    private ObservableMap<String, Double> capacityMap;
    /** Box pro vybírání portu */
    private ComboBox<String> portBox;
    /** Label pro časové info */
    private Label timeLabel;
    /** Tlačítko pro start měření */
    private Button startButton;
    /** Tlačítko pro zastavení měření */
    private Button stopButton;
    /** Tlačítko pro resetováno obsahu grafu a mapy */
    private Button resetButton;
    /** Tlačítko pro uložení obrázku grafu */
    private Button exportChart;
    /** Tlačítko pro uložení měření */
    private Button exportDataButton;
    /** Spinner pro výběr velikosti proudu */
    private Spinner<Integer> voltageSpinner;
    /** Pole pro zadání názvu měření */
    private TextField lineField;
    /** Měření času */
    private Timeline timeline;

    /** Velikost rozestupů prvků */
    private final int spacingSize = 10;
    /** Preferovaná velikost prvků */
    private final int preferredElementSize = 100;
    /** Regulární výraz pro ověření správnosti názvu měření */
    private final String lineNameRegex = "([A-Za-z0-9_]{1,16})";


    /**
     * Metod main. Načítá řetězcové zdroje a inicializuje JavaFx toolkit.
     * @param args - Nepoužito
     */
    public static void main(String[] args) {

        properties = Utility.loadPropertiesFile();
        launch(args);
    }

    /**
     * Metoda inicializující okno aplikace
     * @param stage - Aplikační okno
     */
    @Override
    public void start(Stage stage) {

        this.stage = stage;
        this.capacityMap = FXCollections.observableHashMap();
        manager.setCapacityMap(capacityMap);
        handler.setCapacityMap(capacityMap);

        if (properties == null) {
            showErrorAlert("String resources error",
                    "Failed to load string resources!");

            Platform.exit();
        }

        handler.setProperties(properties);
        stage.setTitle(properties.getProperty("window.title"));

        int windowWidth = 1000;
        int windowHeight = 800;
        stage.setScene(new Scene(getWindowContent(), windowWidth, windowHeight));

        stage.setOnCloseRequest(e -> {
            Platform.exit();
            System.exit(0);
        });

        stage.show();
    }

    /**
     * Vrátí obsah okna aplikace
     * @return - Prvek obsahující veškeré prvky okna aplikace
     */
    private Parent getWindowContent() {

        BorderPane root = new BorderPane();
        int insetsSize = 20;
        root.setPadding(new Insets(insetsSize));
        root.setRight(getControlPanel());
        root.setCenter(getChart());

        stopButton.setDisable(true);
        resetButton.setDisable(true);
        exportChart.setDisable(true);
        exportDataButton.setDisable(true);

        return root;
    }

    /** Vrátí část okna obsahující prvky, které používá uživatel k ovládání aplikace
     *
     * @return - Prvek obsahující orvky ovládání
     */
    private Parent getControlPanel() {

        VBox vbx = new VBox();
        vbx.setSpacing(spacingSize);
        vbx.getChildren().addAll(getCapacityTable(), getTimeInfo(), getPortControls(), getVoltageSpinner(),
                getLineNameField(), getButtons());
        vbx.setAlignment(Pos.CENTER);

        return vbx;
    }

    /**
     * Vrací prvek zobrazující čas
     * @return - Prvek zobrazující čas
     */
    private Node getTimeInfo() {

        timeLabel = new Label(properties.getProperty("time.default"));
        Label descriptionLabel = new Label(properties.getProperty("time.elapsed.label"));

        HBox hbx = new HBox();
        hbx.setAlignment(Pos.CENTER_RIGHT);
        hbx.getChildren().addAll(descriptionLabel, timeLabel);
        hbx.setSpacing(spacingSize);
        return hbx;
    }

    /**
     * Vrací prvek pro zvolení portu
     * @return - Prvek pro vybrání portu
     */
    private Node getPortControls() {

        Label portLabel = new Label(properties.getProperty("port.label"));

        this.portBox = new ComboBox<>();
        this.portBox.setPromptText(properties.getProperty("port.combobox"));
        this.portBox.setMinWidth(preferredElementSize);
        this.portBox.setMaxWidth(preferredElementSize);

        HBox hbx = new HBox();
        hbx.setSpacing(spacingSize);
        hbx.getChildren().addAll(portLabel, this.portBox);
        hbx.setAlignment(Pos.CENTER_RIGHT);

        SerialPort[] ports = SerialPort.getCommPorts();

        if(ports != null)
            for (SerialPort port: ports) {
                portBox.getItems().add(port.getSystemPortName());
            }
        else
            showErrorAlert(properties.getProperty("ports.error.title"),
                    properties.getProperty("ports.error.message"));

        return hbx;
    }

    /**
     * Vrací prvek pro zadání velikosti proudu pro vybíjení
     * @return - Prvek pro vybrání velikosti proudu
     */
    private Node getVoltageSpinner() {

        HBox hbx = new HBox();
        hbx.setSpacing(spacingSize);
        hbx.setAlignment(Pos.CENTER_RIGHT);

        Label voltageLabel = new Label(properties.getProperty("voltage.spinner.label"));
        int spinnerDefaultValue = 20;
        int spinnerMaxValue = 2500;
        int spinnerStep = 100;
        int spinnerMinValue = 20;
        voltageSpinner = new Spinner<>(spinnerMinValue, spinnerMaxValue, spinnerDefaultValue, spinnerStep);
        voltageSpinner.setMaxWidth(preferredElementSize);
        voltageSpinner.setEditable(true);

        hbx.getChildren().addAll(voltageLabel, voltageSpinner);
        return hbx;
    }

    /**
     * Vrací prvek pro zadání názvu měření
     * @return - prvek pro zadání názvu měření
     */
    private Parent getLineNameField() {

        HBox hbx = new HBox();
        hbx.setSpacing(spacingSize);
        hbx.setAlignment(Pos.CENTER_RIGHT);

        Label lineLabel = new Label(properties.getProperty("line.name.label"));

        lineField = new TextField();
        lineField.setMaxWidth(preferredElementSize);
        lineField.setTextFormatter(new TextFormatter<>(c -> {

            if (c.getControlNewText().matches(lineNameRegex) ||
                    c.getControlNewText().matches(""))
                return c;

            return null;
        }));

        hbx.getChildren().addAll(lineLabel, lineField);

        return hbx;
    }

    /**
     * Vrací prvek obsahující všechna tlačítka
     * @return - Prvek obsahující všechna tlačítka
     */
    private Parent getButtons() {

        startButton = new Button(properties.getProperty("start.button"));
        stopButton = new Button(properties.getProperty("stop.button"));
        resetButton = new Button(properties.getProperty("reset.button"));
        exportChart = new Button(properties.getProperty("export.chart.button"));
        Button importDataButton = new Button(properties.getProperty("import.data.button"));
        exportDataButton = new Button(properties.getProperty("export.data.button"));
        Button exitButton = new Button(properties.getProperty("exit.button"));

        startButton.setMinWidth(preferredElementSize);
        stopButton.setMinWidth(preferredElementSize);
        resetButton.setMinWidth(preferredElementSize);
        exportChart.setMinWidth(preferredElementSize);
        importDataButton.setMinWidth(preferredElementSize);
        exportDataButton.setMinWidth(preferredElementSize);
        exitButton.setMinWidth(preferredElementSize);

        startButton.setOnAction(event -> startMeasurement());
        stopButton.setOnAction(event -> stopMeasurement());
        resetButton.setOnAction(event -> resetMeasurement());
        exportChart.setOnAction(event -> exportChart());
        importDataButton.setOnAction(event -> loadData());
        exportDataButton.setOnAction(event -> exportData());
        exitButton.setOnAction(event -> Platform.exit());

        VBox vbx = new VBox();
        vbx.setSpacing(spacingSize);
        vbx.setAlignment(Pos.CENTER);
        vbx.getChildren().addAll(startButton, stopButton, resetButton, exportChart, importDataButton,
                exportDataButton, exitButton);

        return vbx;
    }

    /**
     * Vrací prvek s grafem
     * @return - Prvek s grafem
     */
    private Parent getChart() {

        NumberAxis xAxis = new NumberAxis();
        NumberAxis yAxis = new NumberAxis();

        xAxis.setLabel(properties.getProperty("x.axis"));
        yAxis.setLabel(properties.getProperty("y.axis"));

        this.lineChart = new LineChart<>(xAxis, yAxis);
        this.lineChart.setTitle(properties.getProperty("chart.title"));
        this.lineChart.setCreateSymbols(false);

        return lineChart;
    }

    /**
     * Vrací tabulku kapacit
     * @return - Tabulka kapacit
     */
    private Node getCapacityTable() {
        ObservableList<String> keys = FXCollections.observableArrayList();

       capacityMap.addListener((MapChangeListener.Change<? extends String, ? extends Double> change) -> {
            boolean removed = change.wasRemoved();
            if (removed != change.wasAdded()) {
                if (removed) {
                    keys.remove(change.getKey());
                } else {
                    keys.add(change.getKey());
                }
            }
        });

        final TableView<String> table = new TableView<>(keys);

        TableColumn<String, String> column1 = new TableColumn<>("Measurement name");
        column1.setCellValueFactory(cd -> Bindings.createStringBinding(cd::getValue));

        TableColumn<String, Double> column2 = new TableColumn<>("Capacity [mAh]");
        column2.setCellValueFactory(cd -> Bindings.valueAt(capacityMap, cd.getValue()));

        table.getColumns().setAll(column1, column2);
        table.setColumnResizePolicy(TableView.CONSTRAINED_RESIZE_POLICY);
        return table;
    }

    /**
     * Spouští časovač
     */
    private void startTimer() {
        timeline = new Timeline(
                new KeyFrame(
                        Duration.seconds(1),
                        event ->
                                Platform.runLater(() ->
                                        timeLabel.setText(Utility.getFormattedTime(
                                                Utility.getElapsedTime(manager.getStartTime())
                                                )
                                        )
                                )
                )
        );

        timeline.setCycleCount(Animation.INDEFINITE);
        timeline.play();
    }

    /**
     * Zobrazí chybový dialog
     * @param title - Titulek dialogu
     * @param message - Zpráva dialogu
     */
    private void showErrorAlert(String title, String message) {

        Alert alert = new Alert(AlertType.ERROR);
        alert.setTitle(title);
        alert.setHeaderText(null);
        alert.setContentText(message);

        alert.showAndWait();
    }

    /**
     * Spustí měření. Tato akce se provede při stisku tlačítka.
     */
    private void startMeasurement() {

        if (!isInputValid())
            return;

        if (lineChart.getData().stream().anyMatch(x -> x.getName().equals(lineField.getText()))) {
            showErrorAlert("Adding series error", "Series with this name already exists");
        }

        XYChart.Series<Number, Number> series = new XYChart.Series<>();
        lineChart.getData().add(series);
        series.setName(lineField.getText() + "-" + voltageSpinner.getValue() + "[mA]");
        capacityMap.put(lineField.getText(), 0.0);
        manager.setSeries(series);
        manager.setPort(portBox.getSelectionModel().getSelectedItem());
        manager.setCurrent(voltageSpinner.getValue());
        manager.startMeasurement();
        startButton.setDisable(true);
        stopButton.setDisable(false);
        resetButton.setDisable(false);
        exportChart.setDisable(false);
        exportDataButton.setDisable(false);
        startTimer();
    }

    /**
     * Zastaví měření. Tato akce se provede při stisku tlačítka.
     */
    private void stopMeasurement() {

        manager.stopMeasurement();
        startButton.setDisable(false);
        stopButton.setDisable(true);
       if(timeline != null) this.timeline.stop();
    }

    /**
     * Vymaže všechna naměřená data. Tato akce se provede při stisku tlačítka.
     */
    private void resetMeasurement() {

        timeLabel.setText(properties.getProperty("time.default"));
        this.stopMeasurement();
        this.lineChart.getData().clear();
        resetButton.setDisable(true);
        exportChart.setDisable(true);
        exportDataButton.setDisable(true);
        capacityMap.clear();
    }

    /**
     * Kontroluje validitu vstupu od uživatele před zahájením měření.
     * @return - true pokud jsou vstupy v pořádku
     */
    private boolean isInputValid() {

        if (this.portBox.getSelectionModel().getSelectedItem() == null) {
            showErrorAlert(properties.getProperty("port.error.title"),
                    properties.getProperty("port.error.message"));

            return false;
        }

        if (this.lineField.getText().isBlank()) {
            showErrorAlert(properties.getProperty("line.name.error.title"),
                    properties.getProperty("line.name.error.message"));
            return false;
        }

        return true;
    }

    /**
     * Uloží obrázek grafu.
     */
    private void exportChart() {

        TextInputDialog td = new TextInputDialog();
        td.setHeaderText(properties.getProperty("image.file.ask"));
        Optional<String> result = td.showAndWait();

        if (result.isEmpty() || result.get().isBlank())
            return;

        if (!handler.exportChart(lineChart.snapshot(new SnapshotParameters(), null),
                result.orElse(null)))
            showErrorAlert("export.error.title", "export.error.message");
    }

    /**
     * Uloží naměřená data.
     */
    private void exportData() {

        for (Series<Number, Number> series : lineChart.getData()) {

            File file = new File(System.getProperty("user.dir")
                    + properties.getProperty("data.folder")
                    + series.getName()
                    + properties.getProperty("data.extension"));

            if (!file.getParentFile().mkdirs()) {
                if (!file.getParentFile().exists()) {
                    showErrorAlert(properties.getProperty("directory.creation.error.title"),
                            properties.getProperty("directory.creation.error.message"));
                }
            }

            if (file.exists()) {

                Alert alert = new Alert(AlertType.CONFIRMATION,
                        String.format(properties.getProperty("data.file.exists"), series.getName()),
                        ButtonType.YES, ButtonType.NO, ButtonType.CANCEL);

                alert.showAndWait();

                if (alert.getResult() == ButtonType.CANCEL)
                    return;
                else if (alert.getResult() == ButtonType.NO)
                    continue;

            }

            if (!handler.exportData(file, series))
                showErrorAlert("export.error.title", "export.error.message");
        }
    }

    /**
     * Načte data ze souboru zvoleného uživatelem.
     */
    private void loadData() {

        FileChooser fc = new FileChooser();
        fc.getExtensionFilters().add(new FileChooser.ExtensionFilter(
                properties.getProperty("extension.filter.description"),
                properties.getProperty("extension.filter.extension")));

        File file = fc.showOpenDialog(stage);

        if (file == null)
            return;

        Series<Number, Number> series = handler.loadData(file);

        if (series != null)
            lineChart.getData().add(series);
        else
            showErrorAlert(properties.getProperty("import.error.title"),
                    properties.getProperty("import.error.message"));
        exportDataButton.setDisable(false);
        exportChart.setDisable(false);
        resetButton.setDisable(false);

    }
}